{-# LANGUAGE BlockArguments #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE OverloadedStrings #-}

module Control.Monad.IRTreeSpec where

import Control.Monad.IRTree
import qualified Control.Monad.IRTree as IRTree
import qualified Control.Monad.RTree as RTree
import Data.Bool
import Data.Expr as Expr
import Data.List.NonEmpty (nonEmpty)
import qualified Data.List.NonEmpty as NE
import qualified Data.Map.Strict as Map
import System.IO
import Test.Hspec
import Test.Hspec.Glitter (onGlitter)

spec :: Spec
spec = do
  binsearchSpec
  expsearchSpec
  examplesSpec

binsearchSpec :: Spec
binsearchSpec = describe "binsearch" do
  it "should guess a number between 0 and 10" do
    binsearch (\a -> do pure (a <= 6)) 0 10 `shouldReturn` 6
    binsearch (\a -> do pure (a <= 3)) 0 10 `shouldReturn` 3

expsearchSpec :: Spec
expsearchSpec = describe "expsearch" do
  it "should guess a number between 0 and 10" do
    expsearch (\a -> do pure (a <= 6, 10 - a)) `shouldReturn` (6, True)
    expsearch (\a -> do pure (a <= 3, 10 - a)) `shouldReturn` (3, True)
    expsearch (\a -> do pure (a <= 9, 10 - a)) `shouldReturn` (9, False)
    expsearch (\a -> do pure (a <= 12, 10 - a)) `shouldReturn` (9, False)

examplesSpec :: Spec
examplesSpec = describe "examples" do
  handle "small-opr-expr" $
    Opr (Cnt 1) (Cnt 2)

  handle "small-let-expr" $
    Let "x" (Cnt 1) (Opr (Cnt 2) (Var "x"))

  handle "double-let-expr" $
    Let "x" (Cnt 1) (Let "y" (Cnt 2) (Opr (Var "x") (Var "y")))

  handle "double-overloading-let-expr" $
    Let "x" (Cnt 1) (Let "x" (Cnt 2) (Opr (Var "x") (Var "x")))
 where
  handle str e = describe (str <> " (" <> prettyExprS 0 e ")") do
    let me = runReaderT (Expr.rExpr e) Map.empty

    it "should extract" do
      IRTree.extract me `shouldBe` e

    let re = runReaderT (Expr.rExpr e) Map.empty

    it "should reduce like iinputs" do
      forM_ (RTree.iinputs re) \(ii, e') -> do
        p <- toPredicate ii
        IRTree.reduce (const p) me `shouldReturn` e'

    let
      predicate :: Expr -> IO Bool
      predicate = pure . contains isOpr

    rex <- runIO $ RTree.reduce predicate re

    onGlitter
      ("test/expected/" <> str <> "-ired")
      ( \f -> do
          withFile f WriteMode \h -> do
            let p = debugPredicate' h predicate
            IRTree.reduce p me `shouldReturn` rex
      )
      do
        it "should find an opr normally" \_ -> do
          IRTree.reduce (const predicate) me `shouldReturn` rex

    onGlitter
      ("test/expected/" <> str <> "-ireda")
      ( \f -> do
          withFile f WriteMode \h -> do
            let p = debugPredicate' h predicate
            let (i, l, _) = IRTree.probe me ""
            _ <- p l i
            IRTree.reduceAdaptive p me `shouldReturn` rex
      )
      do
        it "should find an opr addabtively" \_ -> do
          IRTree.reduceAdaptive (const predicate) me `shouldReturn` rex

debugPredicate'
  :: (Show l)
  => Handle
  -> (Expr -> IO Bool)
  -> [AnnotatedChoice l]
  -> Expr
  -> IO Bool
debugPredicate' h predicate lst i = do
  x <- predicate i
  hPutStr
    h
    ( map (debugChoice . choice) lst
        <> ": "
        <> prettyExprS 0 i ""
        <> " "
        <> show x
        <> "\n"
    )
  pure x
 where
  debugChoice = \case
    Yes -> '1'
    No -> '0'
    Undecided -> '*'

debugPredicate
  :: Handle
  -> (l -> ShowS)
  -> (i -> ShowS)
  -> (i -> IO Bool)
  -> [AnnotatedChoice l]
  -> i
  -> IO Bool
debugPredicate h ppl ppi predicate lst i = do
  x <- predicate i
  hPutStr
    h
    ( showString (bool "0" "1" x)
        . showString ": "
        . ppi i
        . showString "\n"
        . case nonEmpty lst of
          Nothing -> showString "initial\n"
          Just lst' -> ppl (label $ NE.last lst') . showString "\n"
        $ ""
    )
  pure x
